"""Declares common exceptions classes."""
import typing

import pydantic
from fastapi.exceptions import RequestValidationError

from unimatrix.exceptions import CanonicalException


class BaseExceptionSpec(pydantic.BaseModel):
    code: str
    message: str
    detail: str
    hint: str

    @classmethod
    def exception_model(cls, impl: typing.Type[CanonicalException]):
        """Create a pydantic model for the given exception class. Must inherit
        from :class:`unimatrix.exceptions.CanonicalException`.
        """
        impl.model = type(f'{impl.__name__}Model', (ExceptionModel,), {
            'kind': pydantic.Field(...,
                title="Kind",
                description=f"Is always equal to `{impl.__name__}`.",
                example=impl.__name__
            ),
            '__annotations__': {
                'kind': str,
                'spec': type(f'{impl.__name__}Spec', (cls,), {
                    'code': pydantic.Field(
                        default=...,
                        title="Error code",
                        example=impl.code
                    ),
                    'message': pydantic.Field(
                        default=...,
                        title="Message",
                        description="A human-readable description of the error condition.",
                        example=impl.message
                    ),
                    'detail': pydantic.Field(
                        default=None,
                        title="Detail",
                        description="A detailed description on why the error condition occurred.",
                        example=impl.detail
                    ),
                    'hint': pydantic.Field(
                        default=None,
                        title="Hint",
                        description="A hint on how to resolve the situation, or an admonition.",
                        example=impl.hint
                    ),
                    'Config': type('Config', (object,), {
                        'title': f'{impl.__name__}Spec'
                    }),
                    '__annotations__': typing.get_type_hints(cls)
                })
            },
            'Config': type('Config', (object,), {
                'title': impl.__name__,
                'summary': impl.__doc__
            })
        })
        return impl

    class Config:
        title = "ExceptionSpec"


exception_model = BaseExceptionSpec.exception_model


class ValidationErrorSpec(BaseExceptionSpec):
    fieldErrors: typing.Dict[str, typing.List[typing.Any]] = {}

    class Config:
        title = "ValidationErrorSpec"


class ExceptionMetadata(pydantic.BaseModel):
    id: str = pydantic.Field(...,
        title="Identifier",
        description=(
            "Identifies the error. May be used to trace an error "
            "instance in logs."
        ),
        example="681b9604-df4e-4b2c-998c-9807a5569b1c"
    )

    timestamp: int = pydantic.Field(...,
        title="Timestamp",
        description=(
            "The date/time at which the error occurred, in milliseconds "
            "since the UNIX epoch."
        ),
        example=1647626707914
    )


class ExceptionModel(pydantic.BaseModel):
    apiVersion: typing.Literal['v1'] = pydantic.Field(...,
        title="API version",
        description="Schema version",
        example="v1"
    )

    metadata: ExceptionMetadata


@exception_model
class UncaughtException(CanonicalException):
    http_status_code: int = 500


@exception_model
class Forbidden(CanonicalException):
    """Raised when the request is unconditionally rejected."""
    code = "FORBIDDEN"
    http_status_code = 403
    message = "The request was unconditionally rejected. That's all we know."


@exception_model
class WebhookMessageRejected(CanonicalException):
    """Raised when an incoming webhook message is rejected."""
    code: str = 'MESSAGE_REJECTED'
    http_status_code: int = 403
    message: str = "The server rejected the message."


@exception_model
class AuthenticationRequired(CanonicalException):
    """Raised when an operation required authenticated requests."""
    code = 'AUTHENTICATION_REQUIRED'
    http_status_code = 401
    message = "The operation requires authenticated requests."
    detail = (
        "An operation against a protected resource was attempted and it does "
        "not allow interactions with unauthenticated requests."
    )


@exception_model
class InvalidAuthorizationScheme(AuthenticationRequired):
    """Raised when the ``Authorization`` header had an invalid scheme."""
    code = 'INVALID_AUTHORIZATION_SCHEME'
    http_status_code = 403
    message = "The scheme provided in the Authorization header is not supported."
    detail = (
        "The server could not parse the credential in the Authorization header "
        "because the specified scheme is unknown or not allowed."
    )


@exception_model
class UntrustedIssuer(AuthenticationRequired):
    """Raised when a token was presented but the issuer is not trusted."""
    code: str = "UNTRUSTED_ISSUER"
    http_status_code: int = 403
    message: str = "The security token was issued by an untrusted party."

    def __init__(self, issuer: str):
        super().__init__(
            detail=(
                "The server has not established a trust relationship with "
                "the issuer ({issuer}) that supplied the security token. "
                "Consult the server documentation on the list of trusted "
                "issuers."
            )
        )


@exception_model
class ForgedAccessToken(AuthenticationRequired):
    """Raised when the parameters of a token were valid but it was not
    issued by a trusted entity.
    """
    code: str = "FORGED_ACCESS_TOKEN"
    http_status_code: int = 403
    message: str = "The security token was valid but not signed by a trusted party."
    detail: str = (
        "The claims presented with the security token validated against the server "
        "policy, but the digital signature did not validate against any of the known "
        "(public) keys. This incident is being logged, including your IP address "
        "and/or a browser fingerprint, and may lead to further escalation, "
        "such as criminal proscecution."
    )
    hint: str = "Stop forging security tokens."


@exception_model
class NotAuthorized(CanonicalException):
    """Raised when a request is not authorized to perform a given operation."""
    code = 'UNAUTHORIZED'
    http_status_code = 403
    message = "This request attempts an unauthorized action."
    detail = (
        "The credentials provided with the request did not resolve to a "
        "principal that was authorized to perform the action."
    )


@exception_model
class BearerAuthenticationRequired(AuthenticationRequired):
    hint = "Provide valid credentials using the Authorization header."


@exception_model
class UpstreamServiceNotAvailable(CanonicalException):
    """Raised when the application is not able to establish
    a (network) connection to an upstream service.
    """
    code = 'SERVICE_NOT_AVAILABLE'
    http_status_code = 503
    message = "The service is currently not available."
    detail = (
        "Network or other infrastructure issues prevent "
        "proper operation of the service."
    )
    hint = "Try again in 600 seconds."


@exception_model
class UpstreamConnectionFailure(UpstreamServiceNotAvailable):
    """Raised when an upstream service listens at the configured address
    and port, but there are issued in establish the connection according
    to the agreed protocol. Such errors may occur when, for example, the
    upstream service is booting and has bound to its address and port, but
    is not yet ready to serve.
    """
    hint = "Try again in 10 seconds."


@exception_model
class TrustIssues(CanonicalException):
    http_status_code = 403
    code = "TRUST_ISSUES"
    message = (
        "The credential attached to the request was issued by an unkown "
        "authority."
    )
    hint = (
        "Verify that the credential (i.e. bearer token or X.509 certificate) "
        "was issued by an authority that is trusted by this system."
    )


@exception_model
class TryAgain(CanonicalException):
    code = 'SERVICE_NOT_AVAILABLE'
    http_status_code = 503
    message = "The service is currently not available."
    detail = (
        "Network or other infrastructure issues prevent "
        "proper operation of the service."
    )
    hint = "Try again in 600 seconds."

    def __init__(self, retry_after): # pragma: no cover
        super().__init__()


@exception_model
class UnsupportedMediaType(CanonicalException):
    http_status_code = 415
    code = "UNSUPPORTED_MEDIA_TYPE"
    message = "The server does not understand the request media type."

    @classmethod
    def fromparsers(cls, parsers: list):
        return cls([x.media_type for x in parsers])

    def __init__(self, accepts: list, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.accepts = accepts

    def get_http_headers(self) -> dict:
        return {
            **super().get_http_headers(),
            'Accept': str.join(', ', self.accepts)
        }


@exception_model
class ParseError(CanonicalException):
    http_status_code = 400
    code = "MALFORMED_BODY"
    message = (
        "The server understands the media type but the request body "
        "is malformed."
    )


@exception_model
class MissingRequestBody(CanonicalException):
    http_status_code: int = 422
    code: str = 'BODY_REQUIRED'
    message: str = "This endpoint requires a request body."
    detail: typing.Optional[str] = None
    hint: typing.Optional[str] = None


@ValidationErrorSpec.exception_model
class UnprocessableEntity(CanonicalException):
    """Test"""
    http_status_code: int = 422
    code: str = "UNPROCESSABLE_ENTITY"
    message: str = "The server was not able to process the request."
    detail: str = (
        "The server understands the content type of the request entity, and "
        "the syntax of the request entity is correct, but it was unable to "
        "process the contained instructions."
    )
    hint: str = "Inspect the .spec.fieldErrors member for additional details."

    @classmethod
    def fromexc(cls, exception: RequestValidationError):
        errors = {}
        for error in exception.errors():
            location = error['loc']
            if error['loc'][0] == 'body' and len(error['loc']) == 1:
                instance = MissingRequestBody()
                break
            else:
                location, field = error['loc']
                if location not in errors:
                    errors[location] = []
        else:
            instance = cls(errors=errors)
        return instance


class Unserializable(ValueError):
    pass


def get_exception_headers(
    status_code: int
) -> typing.Dict[str, typing.Any]:
    return {
        'X-Canonical-Exception': {
            'description': "The name of the exception that occurred.",
            'schema': {
                'type': "string",
                'enum': sorted(set([
                    cls.__name__
                    for cls in CanonicalException.registered
                    if cls.http_status_code == status_code
                ]))
            }
        },
        'X-Error-Code': {
            'description': "The error code describing the exception that occurred.",
            'schema': {
                'type': "string",
                'enum': sorted(set([
                    cls.code
                    for cls in CanonicalException.registered
                    if cls.http_status_code == status_code
                ]))
            }
        },
    }


def get_exception_models(
    status_code: int
) -> typing.TypeVar:
    return typing.Union[tuple(
        cls.model for cls in CanonicalException.registered
        if cls.http_status_code == status_code
        and getattr(cls, 'model', None) is not None
    )]


EXCEPTION_RESPONSES = {
    400: {
        'headers': get_exception_headers(400),
        'description': "The request is malformed",
        'model': get_exception_models(400),
    },
    401: {
        'headers': get_exception_headers(401),
        'description': (
            "Authentication or privilege escalation is required for this endpoint."
        ),
        'model': get_exception_models(401)
    },
    403: {
        'headers': get_exception_headers(403),
        'description': "The request was unconditionally rejected.",
        'model': get_exception_models(403)
    },
    415: {
        'headers': get_exception_headers(415),
        'description': "The server does not understand the request media type.",
        'model': get_exception_models(415)
    },
    422: {
        'headers': get_exception_headers(422),
        'description': "The content of the request was semantically invalid.",
        'model': get_exception_models(422)
    },
    'default': {
        'headers': get_exception_headers(500),
        "description": (
            "An unrecoverable exception occurred while processing the request."
        ),
        'model': get_exception_models(500)
    },
    503: {
        'headers': get_exception_headers(503),
        'description': (
            "The service or an upstream service is not available."
        ),
        'model': get_exception_models(503)
    }
}