"""Declares :class:`Resource`."""
import copy
import inspect
import logging
import typing

import fastapi
import fastapi.params
from fastapi import Request
from fastapi.responses import Response
from fastapi.dependencies.utils import get_dependant
from fastapi.dependencies.utils import get_parameterless_sub_dependant
from fastapi.dependencies.utils import solve_dependencies

from ..bodyconsumer import BodyConsumer
from ..exceptions import AuthenticationRequired
from ..lib.auth import IPrincipal
from ..lib.auth import NullPrincipal
from ..models import BaseEnvelope
from .iresource import IResource
from .resourcemetaclass import ResourceMetaclass
from .resourceoptions import ResourceOptions
from .resourcerequesthandler import ResourceRequestHandler


class Resource(IResource, BodyConsumer, metaclass=ResourceMetaclass):
    """Represents a resource in a web API."""
    __abstract__ = True
    __module__ = 'cbra.resource'
    logger: logging.Logger = logging.getLogger('uvicorn')
    parent: typing.Optional[typing.Type[ResourceMetaclass]] = None
    meta: ResourceOptions

    #: The name of the resource.
    name: str

    #: The name of the resource identifier as a path parameter.
    path_parameter: str

    #: Indicates if authentication is required.
    require_authentication: bool = True

    #: The dependency class used to resolve the current principal.
    principal_class = NullPrincipal

    #: The current principal.
    principal: typing.Any = None

    #: The subresources for this resource.
    subresources: list = []

    #: Additional dependencies specified for this class.
    dependencies: typing.List[fastapi.params.Depends] = []

    #: The resource class definition used by the endpoints.
    resource_class: type = None

    @classmethod
    def add_to_parent(cls, parent: type):
        """Attaches the :class:`cbra.resource.Resource` to a parent resource."""
        child = copy.deepcopy(cls)
        child.meta.parent = parent
        parent.subresources.append(child)

    @classmethod
    def add_to_router(cls,
        app: fastapi.FastAPI,
        base_path: str = '',
        path: str = None
    ) -> None:
        """Add the :class:`Resource` to a FastAPI ASGI application.

        Args:
            app (:class:`fastapi.FastAPI`): the application to add the resource
                to.
            base_path (str): the base path; must start with a slash and not
                end with a slash. Should not contain any path parameters.
            path (str): if not ``None``, use an explicit path instead of the
                :attr:`name` attribute.

        Returns:
            None
        """
        base_path = f'{base_path}/{str.lstrip(path or cls.meta.pluralname, "/")}'
        for _, handler in\
        inspect.getmembers(cls, lambda attr: isinstance(attr, ResourceRequestHandler)):
            handler.add_to_router(app, cls, base_path)
        for sub in cls.subresources:
            sub.add_to_router(
                app=app,
                base_path=f'{base_path}/{{{cls.meta.path_parameter}}}'
            )

    def setup_request_context(
        self,
        method: str,
        path: str,
        request: fastapi.Request,
        response: fastapi.Response,
        principal: typing.Optional[IPrincipal] = None
    ):
        self.method = method
        self.path = path
        self.principal = principal
        self.request = request
        self.response = response

    def get_dependants(self):
        return []

    def is_authenticated(self) -> bool:
        """Return a boolean indicating if the request is authenticated."""
        return bool(self.principal)

    async def authorize(self) -> bool:
        """Hook to perform authorization checks for the request. Must raise
        an appropriate exception if the request is not authorized. This
        method may specify dependencies.
        """
        pass

    async def dispatch(self, handle: typing.Callable):
        if self.require_authentication and not self.is_authenticated():
            raise AuthenticationRequired
        dependant = get_dependant(path=self.path, call=handle)
        dependant.dependencies.extend(self.__get_dependants())
        values, errors, tasks, response, _ =  await solve_dependencies(
            request=self.request,
            dependant=dependant,
            body=await self.get_body(),
            dependency_overrides_provider=None
        )
        if response.status_code is not None: # pragma: no cover
            # TODO
            raise NotImplementedError
        if tasks is not None: # pragma: no cover
            # TODO
            raise NotImplementedError
        if errors: # pragma: no cover
            # Raise the first error
            raise NotImplementedError(errors)
        return await dependant.call(self, **values)

    def update_envelope(self,
        request: fastapi.Request,
        envelope: BaseEnvelope,
        handler: ResourceRequestHandler
    ):
        """Updates the envelope with the relevant properties."""
        if envelope.is_list():
            for resource in envelope.items:
                self.update_envelope(request, resource, handler)

        # Inspect all request handlers defined for the resource class
        # to reverse the proper URLs for the envelope.
        for _, member in inspect.getmembers(self):
            if not isinstance(member, ResourceRequestHandler)\
            or (envelope.is_list() and member.is_detail)\
            or member.action is None:
                continue
            params = {}
            action = member.action

            # TODO: cleanup and refactor
            if member.is_detail and envelope.is_resource():
                params[self.meta.path_parameter] = envelope.get_path_parameter(request)
                envelope.metadata.id = envelope.get_path_parameter(request)
            if envelope.is_list() and action == "list":
                action = "self"
                url = str(self.request.url)
            elif envelope.is_resource() and action == "retrieve":
                action = "self"
                url = member.reverse(self.request, envelope, params)
            else:
                url = member.reverse(self.request, envelope, params)
            envelope.set_link(action, url)

        # Iterate over all subresources and collect the non-detail endpoints,
        # which are relative to the detail URL of the given object.
        if envelope.is_resource():
            params = {
                **self.request.path_params,
                self.meta.path_parameter: envelope.get_path_parameter(request)
            }
            for sub in self.subresources:
                for _, member in inspect.getmembers(sub):
                    if not isinstance(member, ResourceRequestHandler):
                        continue
                    if member.is_detail or not member.action:
                        continue
                    url = member.reverse(self.request, envelope, params)
                    name = f'{sub.meta.pluralname}.{member.action}'
                    if member.action == 'list':
                        name = sub.meta.pluralname
                    envelope.set_link(name, url)

    def __get_dependants(self):
        dependants = self.get_dependants()
        dependants.extend([
            self.authorize
        ])
        return [
            get_parameterless_sub_dependant(
                depends=fastapi.Depends(func),
                path=self.path
            ) for func in dependants
        ]
