"""Declares :class:`OptionsHandler`."""
import typing

import fastapi
import fastapi.responses

from .resourcerequesthandler import ResourceRequestHandler
from .resourceoptions import ResourceOptions


class OptionsHandler(ResourceRequestHandler):
    __module__ = 'cbra.resource'
    description: str  = """Handles options requests to the given base path."""
    response_description: str = "Endpoint options"
    responses: typing.Dict[int, typing.Any] = {
        200: {}
    }

    def __init__(self, allowed_methods: typing.List[str], *args, **kwargs):
        super().__init__(*args, **kwargs)
        allowed_methods.append('OPTIONS')
        self.allowed_methods = list(sorted(set(allowed_methods)))

    def get_router_kwargs(self) -> dict:
        return {
            'response_class': fastapi.Response,
            'responses': {
                200: {
                    'description': "Endpoint options",
                    'model': None,
                    'headers': {
                        'Allow': {
                            'description': "The permitted HTTP methods on this endpoint.",
                            'schema': {
                                'type': "string",
                                'example': "GET,POST"
                            }
                        }
                    },
                }
            }
        }

    def get_summary(self, meta: ResourceOptions) -> typing.Optional[str]:
        """Return the summary of the endpoint."""
        return (
            f"Get the permitted methods on {meta.verbose_name} collections"
            if not self.is_detail else
            f"Get the permitted methods on {meta.verbose_name} objects"
        )

    async def __call__(self,
        request: fastapi.Request,
        response: fastapi.responses.Response,
        principal: dict,
        *args, **kwargs
    ):
        return fastapi.responses.Response(
            headers={
                'Allow': str.join(', ', self.allowed_methods)
            }
        )
