"""Declares :class:`ResourceMetaclass`."""
import abc

from .resourcerequesthandler import ResourceRequestHandler
from .optionshandler import OptionsHandler
from .resourceoptions import ResourceOptions


ACTIONS = [
    ('GET', 'list', False),
    ('POST', 'create', False),
    ('DELETE', 'purge', False),
    ('HEAD', 'exists', True),
    ('GET', 'retrieve', True),
    ('POST', 'receive', True),
    ('PATCH', 'update', True),
    ('PUT', 'replace', True),
    ('DELETE', 'destroy', True),
]


class ResourceMetaclass(abc.ABCMeta):
    __module__ = 'cbra.resource'

    def __new__(cls, name, bases, attrs):
        super_new = super().__new__
        if attrs.pop('__abstract__', False):
            return super_new(cls, name, bases, attrs)

        attrs['meta'] = opts = ResourceOptions.fromclass(
            meta=attrs.pop('Meta', None),
            attrs=attrs,
            resource_class=attrs.pop('resource_class', None)
        )
        subresources = attrs.pop('subresources', [])
        new_class = super_new(cls, name, bases, {**attrs,
            'subresources': []
        })

        # Set the new class as the parent for all subresources. Since
        # subresources may be used by multiple resources, copy the class.
        for subresource in subresources:
            subresource.add_to_parent(new_class)

        # Find all actions and wrap the methods in a RequestHandler
        # instance.
        detail_methods = []
        list_methods = []
        for method, action, is_detail in ACTIONS:
            if action not in attrs:
                continue
            handle = getattr(new_class, action)
            if isinstance(handle, ResourceRequestHandler): # pragma: no cover
                # Nothing to do for manually crafted RequestHandler
                # instances.
                continue
            handler_class = getattr(handle, 'handler_class', ResourceRequestHandler)
            func = handler_class(
                meta=opts,
                action=action,
                method=method,
                handle=handle,
                is_detail=is_detail,
                principal=new_class.principal,
                principal_class=new_class.principal_class
            )
            setattr(new_class, action, func)
            detail_methods.append(method)\
                if is_detail else\
                list_methods.append(method)

        if detail_methods:
            setattr(new_class, '_options_detail', OptionsHandler(
                allowed_methods=detail_methods,
                meta=opts,
                action=None,
                method='OPTIONS',
                handle=None,
                is_detail=True,
            ))
        if list_methods:
            setattr(new_class, '_options_list', OptionsHandler(
                allowed_methods=list_methods,
                meta=opts,
                action=None,
                method='OPTIONS',
                handle=None,
                is_detail=False,
            ))

        return new_class
