"""Declares :class:`ResourceOptions`."""
import typing
from inspect import Parameter

import fastapi
import inflect
import pydantic
from unimatrix.exceptions import ImproperlyConfigured

from ..models import BaseEnvelope
from ..models import EnvelopeMetadata


DETAIL_ACTIONS = {'receive', 'retrieve', 'update', 'delete', 'replace'}


class ResourceOptions:
    """The :class:`cbra.resource.ResourceOptions` is a base class for the
    configured options of a :class:`cbra.Resource` implementation.
    """
    __module__: str = 'cbra.resource'
    parent: type = None

    #: The singular name of the resource. This is a required attribute on the
    #: inner ``Meta`` class of a :class:`cbra.Resource`."""
    name: str

    #: The plural name of the resource. This attribute is optional, and if not
    #: specified, one is generated from :attr:`name`.
    pluralname: typing.Optional[str] = None

    #: Specifies the name of the path parameter in URLs. If :attr:`path_parameter`
    #: is ``'user_id'``, then the URL constructed for the ``users`` resource
    #: becomes ``'/users/{user_id}'``. The default value is ``'resource_id'``.
    #:
    #: If the :attr:`path_parameter` attribute has an annotation, this is also
    #: used to coerce the path parameter in the given type.
    path_parameter: str

    #: The :class:`fastapi.Path` instance representing the path parameter of
    #: the resource.
    path_dependency: fastapi.Path

    #: The default value for the path parameter. An :class:`Ellipsis` indicates
    #: that the parameter is mandatory.
    path_default: typing.Any = ...

    #: The name of the object returned from the endpoint.
    resource_name: str = None

    #: The default version of the object schema.
    resource_version: str = 'v1'

    #: The list of tags to group the resource under in the UI.
    tags: typing.List[str]

    #: The human readable name of the resource.
    verbose_name: str = None

    #: Like :attr:`verbose_name`, but plural.
    verbose_name_plural: str = None

    @staticmethod
    def has_detail_methods(attrs: dict) -> bool: # pragma: no cover
        return bool(set(attrs.keys()) & DETAIL_ACTIONS)

    @staticmethod
    def has_subresources(attrs: dict) -> bool: # pragma: no cover
        return bool(attrs.get('subresources'))

    @classmethod
    def fromclass(cls, meta: typing.Optional[type], *args, **kwargs):
        bases = [cls]
        if meta is not None:
            bases.insert(0, meta)
        new_class = type('ResourceOptions', tuple(bases), {})
        return new_class(*args, **kwargs)

    def __init__(self, attrs: dict, resource_class: typing.Optional[type] = None):
        hints = typing.get_type_hints(self)
        if not hasattr(self, 'path_parameter')\
        and (self.has_detail_methods(attrs) or self.has_subresources(attrs)): # pragma: no cover
            raise ImproperlyConfigured(
                "The Meta.path_parameter attribute is required."
            )
        self.path_dependency = fastapi.Path(
            default=self.path_default,
            alias=self.path_parameter
        )
        self.path_signature = Parameter(
            name=self.path_parameter,
            kind=Parameter.POSITIONAL_ONLY,
            annotation=hints.get(self.path_parameter) or str
        )

        # Generate a plural name if one is not specified on the inner Meta
        # class.
        if self.pluralname is None:
            engine = inflect.engine()
            self.pluralname = engine.plural(self.name)
        assert self.name != self.pluralname # nosec
        if self.verbose_name is None:
            self.verbose_name = str.title(self.name)
        if self.verbose_name_plural is None:
            self.verbose_name_plural = str.title(self.pluralname)
        self.tags = [str.title(self.name)]

        # Create envelopes for a single resource as well as a collection.
        self.envelope_model = self.list_model = self.spec_model = None
        self.resource_name = self.resource_config = None
        if resource_class is not None:
            self.resource_config = getattr(resource_class, 'Config', None)
            self.resource_name = getattr(self.resource_config, 'title', None)\
                or str.title(self.resource_name or self.name)
            self.envelope_model, self.list_model, self.spec_model = (
                self.create_envelope_model(resource_class)
            )

    def create_envelope_model(self, cls: type) -> pydantic.BaseModel:
        """Creates the envelope for the resource defined by the resource
        class attribute.
        """
        name = self.resource_name

        # Create the spec model based on the defined parameters.
        Spec = cls
        ResourceEnvelope = type(name, (BaseEnvelope,), {
            'apiVersion': pydantic.Field(...,
                title="API version",
                description=f"Is always `{self.resource_version}`.",
                example=self.resource_version
            ),
            'kind': pydantic.Field(...,
                title="Kind",
                description=f"Is always `{name}`.",
                example=name
            ),
            'metadata': pydantic.Field(
                title="Metadata",
                description="Specification of the object metadata.",
                default_factory=lambda: EnvelopeMetadata()
            ),
            #'spec': pydantic.Field(...,
            #    title="Specification",
            #    description=f"Specification of the {name} resource."
            #),
            'Config': type('Config', (self.resource_config or object,), {
                'title': f'{name}Envelope'
            }),
            '_is_resource': True,
            '_is_list': False,
            '__annotations__': {
                'apiVersion': str,
                'kind': str,
                'metadata': EnvelopeMetadata,
                'spec': Spec
            }
        })

        # Create the list with similar parameters
        ListEnvelope = type(f'{name}List', (BaseEnvelope,), {
            'apiVersion': pydantic.Field(...,
                title="API version",
                description=f"Is always `{self.resource_version}`.",
                example=self.resource_version
            ),
            'kind': pydantic.Field(...,
                title="Kind",
                description=f"Is always `{name}List`.",
                example=f'{name}List'
            ),
            'metadata': pydantic.Field(
                title="Metadata",
                description="Specification of the list metadata.",
                default_factory=lambda: EnvelopeMetadata()
            ),
            'items': pydantic.Field(...,
                title="Items",
                description=f"Items in the list of {name} objects."
            ),
            'Config': type('Config', (object,), {
                'title': f'{name}List'
            }),
            '_envelope_factory': lambda result: ResourceEnvelope(
                apiVersion=self.resource_version,
                kind=name,
                spec=result
            ),
            '_is_resource': False,
            '_is_list': True,
            '__annotations__': {
                'apiVersion': str,
                'kind': str,
                'metadata': EnvelopeMetadata,
                'items': typing.List[ResourceEnvelope]
            }
        })

        return ResourceEnvelope, ListEnvelope, Spec

    def get_envelope_name(self, detail: bool) -> str:
        """Return a string holding the ``kind`` attribute of the top-level
        envelope.
        """
        return self.resource_name\
            if detail else\
            f'{self.resource_name}List'

    def get_mount_path(self, base_path: str, detail: bool = False) -> str:
        """Return a string holding the base path to mount the resource
        on.
        """
        path = base_path
        if detail:
            path = f'{path}/{{{self.path_parameter}}}'
        return path

    def get_tags(self, parent: typing.Optional[type]) -> typing.List[str]:
        """Return the list of tags to register the endpoint under."""
        return self.tags if parent is None\
            else parent.meta.get_tags(getattr(parent, 'parent', None))

    def get_path_signature(self, detail: bool = False) -> typing.List[Parameter]:
        """Return a list of :class:`inspect.Signature` instance representing
        the path parameters of a subresources' parents and its own path.
        """
        params: typing.List[Parameter] = []
        cls = self.parent
        while cls is not None:
            params.append(cls.meta.path_signature)
            cls = cls.meta.parent
        params = list(reversed(params))
        if detail:
            params.append(self.path_signature)
        return params
