"""Declares :class:`RequestHandler`."""
import asyncio
from re import A
import typing
import inspect
from cbra.resource.iresource import IResource

import fastapi
import fastapi.responses
import pydantic

from ..bodyconsumer import BodyConsumer
from ..dependantsignature import DependantSignature
from ..handler import BaseRequestHandler
from ..models import BaseEnvelope
from .ienvelope import IEnvelope
from .iresource import IResource
from .resourceoptions import ResourceOptions


class ResourceRequestHandler(BaseRequestHandler, BodyConsumer):
    """Handles a HTTP request to an endpoint, for a specific request method."""
    _is_coroutine = asyncio.coroutines._is_coroutine
    __module__ = 'cbra.resource'
    meta: ResourceOptions

    @property
    def __doc__(self) -> str:
        return self.handle.__doc__

    def __init__(self,
        meta: ResourceOptions,
        action: str,
        method: str,
        handle: typing.Callable,
        is_detail: bool,
        **kwargs
    ):
        # Check if the asyncio.iscoroutinefunction() call returns
        # True for this object, since it depends on a private
        # symbol.
        assert asyncio.iscoroutinefunction(self) # nosec
        self.meta = meta
        self.action = action
        self.handle = handle
        self.is_detail = is_detail
        self.path = None

        returns = None
        if self.handle is not None:
            # Remove the `self` argument from the signature because it confuses
            # FastAPI (it sees it as a query parameter).
            sig = inspect.signature(self.handle)

            # Only add the return annotation if it is a pydantic.BaseModel
            # subclass; other types cause issues (TODO).
            if inspect.isclass(sig.return_annotation)\
            and issubclass(sig.return_annotation, pydantic.BaseModel)\
            or isinstance(sig.return_annotation, typing._UnionGenericAlias):
                returns = sig.return_annotation
            self.handle.__signature__ = sig.replace(
                parameters=[
                    x for x in sig.parameters.values()
                    if x.name != 'self'
                ]
            )

        super().__init__(
            method=method,
            returns=returns,
            **kwargs
        )

    def add_to_router(self,
        app: fastapi.FastAPI,
        cls: type,
        base_path: str
    ) -> None:
        """Add the :class:`RequestHandler` to a FastAPI ASGI application.

        Args:
            app (:class:`fastapi.FastAPI`): the application to add the resource
                to.
            cls (:class:`cbra.Resource`): the resource class.
            base_path (str): the base path; must start with a slash and not
                end with a slash. Should not contain any path parameters.

        Returns:
            None
        """
        self.meta = cls.meta
        self.name=(
            f'{self.meta.pluralname}.{str.lower(self.action)}'
            if self.action is not None else None
        )
        self.resource_class = cls
        return super().add_to_router(
            app=app,
            name=self.name,
            path=self.meta.get_mount_path(
                base_path=base_path,
                detail=self.is_detail
            ),
            response_model=self.get_response_model(),
            summary=self.get_summary(cls.meta),
            tags=self.meta.get_tags(parent=cls.meta.parent),
        )

    def get_handler_kwargs(
        self,
        *args: typing.Any,
        **kwargs: typing.Any
    ) -> typing.Dict[str, typing.Any]:
        """Return the keyword arguments that are passed to the request
        handler function.
        """
        return kwargs

    def get_response_model(
        self
    ) -> typing.Union[IEnvelope, typing.Any]:
        """Return the :class:`pydantic.BaseModel` implementation to serialize
        the response.
        """
        # Allow overriding of the return value with the return type
        # annotation in the request handler.
        if self.returns:
            return self.returns

        model = None
        if self.is_envelope_response():
            model = self.meta.envelope_model
        elif self.is_list_response():
            model = self.meta.list_model
        return model

    def get_signature_parameters(self) -> typing.List[inspect.Parameter]:
        """Get the list of :class:`inspect.Parameter` instances specifying
        the callables' signature.
        """
        params = []
        if self.handle:
            params.extend(inspect.signature(self.handle).parameters.values())
        params.extend([
            inspect.Parameter(
                name='handler',
                kind=inspect.Parameter.KEYWORD_ONLY,
                annotation=self.resource_class,
                default=fastapi.Depends()
            )
        ])
        return params

    def get_method_handler(
        self,
        resource: IResource
    ) -> typing.Callable[..., IResource.ReturnTypes]:
        """Return the handle function for the current request method."""
        return self.handle

    def get_summary(self, meta: ResourceOptions) -> typing.Optional[str]:
        """Return the summary of the endpoint."""
        if self.action == 'exists':
            return f"Check if {meta.verbose_name} exists"

        return (
            f'{str.title(self.action)} {meta.verbose_name}'
            if self.is_detail or self.action == 'create' else
            f'{str.title(self.action)} {meta.verbose_name_plural}'
        )

    def is_envelope_response(self) -> bool:
        """Return a boolean indicating if the response expects an envelope."""
        return self.action in {"create", "replace", "retrieve", "update"}

    def is_list_response(self) -> bool:
        """Return a boolean indicating if the response expects a list."""
        return self.action in {"list"}

    def reverse(self,
        request: fastapi.Request,
        envelope: BaseEnvelope,
        params: typing.Dict[str, typing.Union[int, str]]
    ) -> str:
        """Reverse the URL of this handler given the path parameters."""
        params = {
            **{k: v for k, v in dict.items(request.path_params)},
            **params
        }
        if not self.is_detail:
            params.pop(self.meta.path_parameter, None)
        return request.url_for(self.name, **params)

    async def run_handle(self,
        request: fastapi.Request,
        response: fastapi.responses.Response,
        principal: typing.Any,
        handler: IResource,
    ) -> typing.Any:
        """Run invoke the handler method corresponding to the request method
        and path.
        """
        handler.setup_request_context(
            method=self.method,
            path=self.path,
            principal=principal,
            request=request,
            response=response
        )
        result = await handler.dispatch(
            handle=self.get_method_handler(handler)
        )
        Model = self.get_response_model()
        if isinstance(result, (dict, list))\
        and handler.resource_class is not None:
            result = Model.fromresult(request, result, {
                'apiVersion': self.meta.resource_version,
                'kind': self.meta.get_envelope_name(self.is_detail)
            })
            handler.update_envelope(request, result, self)
        elif isinstance(result, pydantic.BaseModel):
            result = result.dict()
        return result
