# pylint: skip-file
import asyncio
import datetime
import json
import secrets

import cbra
import ckms
import ckms.jose
from cbra.ext import oauth2
from cbra.ext.oauth2 import MemoryClientRepository
from cbra.ext.oauth2 import MemorySubjectRepository
from cbra.ext.oauth2 import MemoryStorage
from cbra.ext.oauth2 import ref


# This is the key of client application. In a real-world implementation
# the public key would be received dynamically.
app = cbra.Application()

MemoryClientRepository.add(ref.clients.jwt)
MemorySubjectRepository.add(ref.alice)
MemorySubjectRepository.add(ref.bob)


# Instantiate the OAuth 2.0 server and add it to the
# ASGI application.
server = oauth2.OpenAuthorizationServer(
    storage=MemoryStorage,
    keychain=ref.keychain,
    token=oauth2.TokenRequestHandler,
    clients=MemoryClientRepository,
    subjects=MemorySubjectRepository
)
server.add_to_app(app)


@app.on_event('startup')
async def print_jwt():
    # Use this JSON Web Signature (JWS) to exchange a token.
    bob = ckms.Keychain()
    bob.register({
        'sig': {
            'provider': 'local',
            'key': ref.keys.bob.get_impl()
        }
    })
    await bob.setup()

    codec = ckms.jose.PayloadCodec(signer=bob)
    now = int(datetime.datetime.utcnow().timestamp())
    jwt = ckms.jose.JSONWebToken(
        iss=ref.bob.sub,
        aud="https://localhost:8000/token",
        sub=ref.bob.sub,
        iat=now,
        exp=now+600,
        nbf=now,
        jti=secrets.token_urlsafe(24)
    )
    jws = await codec.encode(jwt).sign(algorithm='EdDSA', using='sig', claims={'kid': 'sig'})
    token = bytes.decode(bytes(jws))
    request = {
        'client_id': "jwt",
        'grant_type': oauth2.GrantType.jwt_bearer.value,
        'assertion': token,
        'scope': "read write"
    }
    print(json.dumps(request, indent=2))


if __name__ == '__main__':
    cbra.run('__main__:app', reload=True)
