# pylint: skip-file
import datetime
from typing import Optional

import uvicorn
import pydantic
from fastapi import Request
from fastapi import Header
from fastapi.responses import Response

import cbra
from cbra.ext import webhook


class WooCommerceObject(pydantic.BaseModel):
    # Some common fields in the WooCommerce API.
    id: int
    date_created: datetime.datetime
    date_created_gmt: datetime.datetime
    date_modified: datetime.datetime
    date_modified_gmt: datetime.datetime


class WoocommerceWebhookHandler(webhook.WebhookHandler):
    """Handles webhook events from WooCommerce."""
    name: str = 'woocommerce'
    summary: str = "WooCommerce"
    allowed_events: set = {
        "order.created",
        "order.updated",
        "order.deleted",
        "product.created",
        "product.updated",
        "product.deleted",
        "product.restored"
    }

    def get_event_name(self,
        topic: Optional[str] = Header(None, alias='X-WC-Webhook-Topic')
    ) -> str:
        """Determine the event based on the `X-WC-Webhook-Topic` header."""
        return topic

    async def handle(self,
        request: Request,
        dto: WooCommerceObject,
        signature: Optional[str] = Header(None, alias='X-Wc-Webhook-Signature'),
        source: Optional[str] = Header(None, alias='X-Wc-Webhook-Source'),
    ):
        """Convert a WooCommerce event that is delivered through a webhook
        to a common, internal event format.
        """
        print(repr(dto))

    async def on_dropped(self,
        request: Request,
        event: str
    ):
        """Handle a dropped message. Messages for this handler are dropped
        when the event type could not be determined based on the value of the
        ``X-WC-Webhook-Topic`` header.

        If the header was not set, this was a ping from WooCommerce; otherwise
        there is a webhook configured for events that this handler is not
        interested in.
        """
        if event is not None:
            response.status_code = 403

    async def verify_message(self):
        """Hook to implement message verification. This method is expected
        for raise a fatal exception if verification fails.
        """
        pass


app = cbra.Application(redoc_url='/docs')
app.add(WoocommerceWebhookHandler(), path='/webhooks/woocommerce')



if __name__ == '__main__':
    uvicorn.run(app)
